/****************************************************************************
    
    jackpluginwrapper.cpp - a class that should make it easier to write 
                            DSSI hosts in C++
    
    Copyright (C) 2005  Lars Luthman <larsl@users.sourceforge.net>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 01222-1307  USA

****************************************************************************/

#include "jackpluginwrapper.hpp"


using namespace std;


JACKPluginWrapper::JACKPluginWrapper(const string& identifier) 
  : PluginWrapper(identifier), 
    m_client(NULL), 
    m_program_selected(false),
    m_alsa_encoder(NULL),
    m_current_bank(0) {
  m_controls.resize(port_count(), NULL);
  snd_midi_event_new(1024, &m_alsa_encoder);
}


JACKPluginWrapper::~JACKPluginWrapper() {
  snd_midi_event_free(m_alsa_encoder);
}


void JACKPluginWrapper::set_client(jack_client_t* client) {
  m_client = client;
}


void JACKPluginWrapper::activate() {
  PluginWrapper::activate();
  
  // don't do the rest if we don't have a client
  if (!m_client)
    return;
  
  // set up ports
  for (unsigned i = 0; i < port_count(); ++i) {
    
    LADSPA_PortDescriptor desc = port_descriptor(i);
    
    // control port
    if (desc & LADSPA_PORT_CONTROL) {
      m_controls[i] = new LADSPA_Data;
      connect_port(i, m_controls[i]);
    }
    
    // audio input port
    else if ((desc & LADSPA_PORT_AUDIO) && (desc & LADSPA_PORT_INPUT)) {
      jack_port_t* port = jack_port_register(m_client, port_name(i),
					     JACK_DEFAULT_AUDIO_TYPE,
					     JackPortIsInput, 0);
      m_jack_input_ports.push_back(PortMap(port, i));
    }
    
    // audio output port
    else if ((desc & LADSPA_PORT_AUDIO) && (desc & LADSPA_PORT_OUTPUT)) {
      jack_port_t* port = jack_port_register(m_client, port_name(i),
					     JACK_DEFAULT_AUDIO_TYPE,
					     JackPortIsOutput, 0);
      m_jack_output_ports.push_back(PortMap(port, i));
    }
  }
  
  // the midi port
  m_midi_input_port = jack_port_register(m_client, "MIDI input", 
					 JACK_DEFAULT_MIDI_TYPE, 
					 JackPortIsInput, 0);
}


void JACKPluginWrapper::run(unsigned long frames) {
  
  // if we don't have a client, just run PluginWrapper::run()
  if (!m_client) {
    PluginWrapper::run(frames);
    return;
  }
  
  // select program if it hasn't been done already
  if (!m_program_selected) {
    // XXX bad, bad, bad
    select_program(0, 0);
    m_program_selected = true;
  }
  
  // fill the MIDI buffer for the plugin
  // XXX This is nasty, should use the ALSA encoder instead
  void* midi_port_buf = jack_port_get_buffer(m_midi_input_port, frames);
  jack_default_midi_event_t midi_event;
  jack_nframes_t event_index = 0;
  jack_nframes_t event_count = 
    jack_midi_port_get_info(midi_port_buf, frames)->event_count;
  int j = 0;
  for (unsigned i = 0; i < event_count && j < 1000; ++i) {
    jack_midi_get_event_n(&midi_event, midi_port_buf, event_index, frames);
    ++event_index;
    
    snd_midi_event_encode(m_alsa_encoder, midi_event.buffer, 
			  midi_event.size, &m_midi_events[j]);
    if (m_midi_events[j].type != SND_SEQ_EVENT_NONE) {
      
      switch (m_midi_events[j].type) {
	
	// note on events with zero velocity should be note off events
      case SND_SEQ_EVENT_NOTEON:
	if (m_midi_events[j].data.note.velocity == 0)
	  m_midi_events[j].type = SND_SEQ_EVENT_NOTEOFF;
	++j;
	break;
	
	// program change events should not be passed to the plugin
      case SND_SEQ_EVENT_PGMCHANGE:
	select_program(m_current_bank, m_midi_events[j].data.control.value);
	break;
	
	// neither should controller events that are mapped to ports
	// or bank select events
      case SND_SEQ_EVENT_CONTROLLER:
	if (m_midi_events[j].data.control.param == 0) {
	  m_current_bank = m_midi_events[j].data.control.value * 128 + 
	    (m_current_bank % 128);
	  break;
	}
	if (m_midi_events[j].data.control.param == 32) {
	  m_current_bank = m_current_bank - (m_current_bank % 128) + 
	    m_midi_events[j].data.control.value;
	  break;
	}
	unsigned long mapped_port =
	  midi_map(m_midi_events[j].data.control.param);
	if (mapped_port != ULONG_MAX) {
	  const LADSPA_PortRangeHint& hint = port_range_hint(mapped_port);
	  if ((hint.HintDescriptor & LADSPA_HINT_BOUNDED_BELOW) &&
	      (hint.HintDescriptor & LADSPA_HINT_BOUNDED_ABOVE)) {
	    signed int& value = m_midi_events[j].data.control.value;
	    *m_controls[mapped_port] = hint.LowerBound + 
	      value * (hint.UpperBound - hint.LowerBound) / 127.0;
	  }
	  break;
	}
	
	// everything else should
      default:
	++j;

      }
    }
  }
  
  // connect JACK port buffers to the plugin ports
  int jack_port = 0;
  for (unsigned i = 0; i < m_jack_input_ports.size(); ++i) {
    jack_port_t* port = m_jack_input_ports[jack_port].jack;
    connect_port(m_jack_input_ports[jack_port].plugin,
		 (LADSPA_Data*)jack_port_get_buffer(port, frames));
    ++jack_port;
  }
  jack_port = 0;
  for (unsigned i = 0; i < m_jack_output_ports.size(); ++i) {
    jack_port_t* port = m_jack_output_ports[jack_port].jack;
    connect_port(m_jack_output_ports[jack_port].plugin,
		 (LADSPA_Data*)jack_port_get_buffer(port, frames));
    ++jack_port;
  }
  
  // execute the plugin callback
  run_synth(frames, m_midi_events, j);
}
