/****************************************************************************
    
    pluginwrapper.cpp - a class that should make it easier to write 
                        DSSI hosts in C++
    
    Copyright (C) 2005  Lars Luthman <larsl@users.sourceforge.net>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 01222-1307  USA

****************************************************************************/

#include <climits>
#include <iostream>
#include <sstream>
#include <stdexcept>

#include <glibmm.h>

#include "pluginwrapper.hpp"


using namespace Glib;
using namespace std;


PluginWrapper::PluginWrapper(const string& identifier) throw(runtime_error)
  : m_library(NULL),
    m_handle(NULL),
    m_dssi(NULL),
    m_activated(false) {
  
  // ULONG_MAX means that a MIDI controller is unmapped (let's hope that
  // nobody uses that many ports!)
  m_midi_map.resize(128, ULONG_MAX);
  
  // get the type, library and label
  istringstream iss(identifier);
  string type, library, label;
  getline(iss, type, ':');
  getline(iss, library, ':');
  getline(iss, label, ':');

  // load the plugin
  runtime_error up("Failed to load plugin");
  bool loaded = false;
  if (type == "DSSI")
    loaded = load_dssi(library, label);
  
  // if we failed, let the caller know
  if (!loaded)
    throw up;
}


PluginWrapper::~PluginWrapper() {
  cout<<"Unloading plugin"<<endl;
  if (m_activated)
    deactivate();
  if (m_handle)
    m_dssi->LADSPA_Plugin->cleanup(m_handle);
  delete m_library;
}


void PluginWrapper::activate() {
  assert(m_handle);
  if (m_dssi->LADSPA_Plugin->activate)
    m_dssi->LADSPA_Plugin->activate(m_handle);
  m_activated = true;
}


void PluginWrapper::deactivate() {
  assert(m_handle);
  if (m_dssi->LADSPA_Plugin->deactivate)
    m_dssi->LADSPA_Plugin->deactivate(m_handle);
  m_activated = false;
}


char* PluginWrapper::configure(const string& key, const string& value) {
  assert(m_handle);
  if (m_dssi->configure)
    return m_dssi->configure(m_handle, key.c_str(), value.c_str());
  return NULL;
}


int PluginWrapper::get_midi_controller_for_port(unsigned long port) {
  assert(m_handle);
  if (m_dssi->get_midi_controller_for_port)
    return m_dssi->get_midi_controller_for_port(m_handle, port);
  return DSSI_NONE;
}


const DSSI_Program_Descriptor* PluginWrapper::get_program(unsigned long index) {
  assert(m_handle);
  if (m_dssi->get_program)
    return m_dssi->get_program(m_handle, index);
  return NULL;
}


void PluginWrapper::connect_port(unsigned long port, LADSPA_Data* buffer) {
  assert(m_handle);
  if (m_dssi->LADSPA_Plugin->connect_port)
    m_dssi->LADSPA_Plugin->connect_port(m_handle, port, buffer);
}


void PluginWrapper::select_program(unsigned long bank, unsigned long program) {
  assert(m_handle);
  if (m_dssi->select_program)
    m_dssi->select_program(m_handle, bank, program);
}


void PluginWrapper::run(unsigned long frames) {
  assert(m_handle);
  if (m_dssi->LADSPA_Plugin->run)
    m_dssi->LADSPA_Plugin->run(m_handle, frames);
}


void PluginWrapper::run_synth(unsigned long framecount, 
			      snd_seq_event_t* events, 
			      unsigned long eventcount) {
  assert(m_handle);
  if (m_dssi->run_synth)
    m_dssi->run_synth(m_handle, framecount, events, eventcount);
  else if (m_dssi->LADSPA_Plugin->run)
    m_dssi->LADSPA_Plugin->run(m_handle, framecount);
}


const char* PluginWrapper::label() const {
  return m_dssi->LADSPA_Plugin->Label;
}


const LADSPA_Properties& PluginWrapper::properties() const {
  return m_dssi->LADSPA_Plugin->Properties;
}


const char* PluginWrapper::name() const {
  return m_dssi->LADSPA_Plugin->Name;
}


const char* PluginWrapper::maker() const {
  return m_dssi->LADSPA_Plugin->Maker;
}


const char* PluginWrapper::copyright() const {
  return m_dssi->LADSPA_Plugin->Copyright;
}


unsigned long PluginWrapper::port_count() const {
  return m_dssi->LADSPA_Plugin->PortCount;
}


const LADSPA_PortDescriptor& 
PluginWrapper::port_descriptor(unsigned long port) const {
  return m_dssi->LADSPA_Plugin->PortDescriptors[port];
}


const char* PluginWrapper::port_name(unsigned long port) const {
  return m_dssi->LADSPA_Plugin->PortNames[port];
}


const LADSPA_PortRangeHint& 
PluginWrapper::port_range_hint(unsigned long port) const {
  return m_dssi->LADSPA_Plugin->PortRangeHints[port];
}


void PluginWrapper::init_midi_map() {
  for (unsigned long i = 0; i < port_count(); ++i) {
    int n;
    if (DSSI_IS_CC(n = get_midi_controller_for_port(i)))
      m_midi_map[DSSI_CC_NUMBER(n)] = i;
  }
}


bool PluginWrapper::load_dssi(const std::string& library, 
			      const std::string& label) {
  
  // get the DSSI path
  string pathlist = getenv("DSSI_PATH");
  if (pathlist == "") {
    pathlist = "/usr/lib/.dssi:/usr/local/lib/.dssi";
    pathlist = pathlist + ":" + getenv("HOME") + "/.dssi";
    cout<<"No DSSI_PATH defined, using "<<pathlist<<endl;
  }
  
  // if the library starts with a /, add the empty directory to the list
  // (so we try to use the library name as an absolute path)
  if (library.size() > 0 && library[0] == '/')
    pathlist = string(":") + pathlist;
  
  // iterate over all directories in the path list
  int n = 0;
  while (n < pathlist.size()) {
    int p = pathlist.find(":", n);
    p = (p == string::npos ? pathlist.size() : p);
    string path = pathlist.substr(n, p - n);
    n = p + 1;
    
    // try to load the library file in this directory
    string name;
    if (path.size() > 0)
      name = path + "/" + library;
    else
      name = library;
    cout<<"Trying to load "<<name<<"... ";
    m_library = new Module(name);
    
    // was the library loadable?
    if (*m_library) {
      void* symbol;
      
      // is the library a DSSI library?
      if (m_library->get_symbol("dssi_descriptor", symbol)) {
	DSSI_Descriptor_Function dssi_descriptor = 
	  (DSSI_Descriptor_Function)symbol;
	
	// iterate over the available plugins in the library
	for (int i = 0; true; ++i) {
	  m_dssi = dssi_descriptor(i);
	  if (!m_dssi)
	    break;
	  const LADSPA_Descriptor* ladspa = m_dssi->LADSPA_Plugin;
	  
	  // is this the plugin we're looking for?
	  if (label == ladspa->Label) {
	    m_handle = 
	      ladspa->instantiate((const LADSPA_Descriptor*)ladspa, 48000);
	    
	    // instantiate it and initialise the MIDI CC mappings
	    if (m_handle) {
	      cout<<"loaded "<<ladspa->Label<<endl;
	      init_midi_map();
	      return true;
	    }
	  }
	}
      }
      else
	cout<<"not a DSSI plugin library, ";
    }
    delete m_library;
    cout<<"failed"<<endl;
    m_library = NULL;
  }
  
  return false;
}
