/****************************************************************************
    
    pluginwrapper.hpp - a class that should make it easier to write 
                        DSSI hosts in C++
    
    Copyright (C) 2005  Lars Luthman <larsl@users.sourceforge.net>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 01222-1307  USA

****************************************************************************/

#ifndef PLUGINWRAPPER_HPP
#define PLUGINWRAPPER_HPP

#include <stdexcept>
#include <string>
#include <vector>

#include <glibmm.h>
#include <dssi.h>



/** This class wraps an instance of a DSSI plugin. It loads the plugin,
    initialises MIDI controller mappings, and provides member functions
    for most of the DSSI callbacks and properties.
*/
class PluginWrapper {
public:
  
  /** The parameter @c identifier should be a string of the format 
      @c "DSSI:libraryname:label", for example 
      @c "DSSI:sineshaper.so:ll-sineshaper". All the directories in the
      environment variable DSSI_PATH will be searched.
  */
  PluginWrapper(const std::string& identifier) throw(std::runtime_error);
  
  ~PluginWrapper();
  
  // XXX Fix the problems mentioned below
  /** This function returns the contol port number that is mapped to
      the given MIDI CC number. If nothing is mapped to the CC this function
      return ULONG_MAX. This function has some pretty severe limitations, 
      for example NRPNs aren't handled at all and you can only map one port
      to every CC number. This has to be fixed if this class is to be used in
      a real DSSI host. */
  unsigned long midi_map(unsigned char controller) const { 
    assert(controller < 128);
    return m_midi_map[controller]; 
  }
  
  /* Here follow wrappers for DSSI callback functions. It should be pretty
     obvious what they do. */
  void activate();
  void deactivate();
  char* configure(const std::string& key, const std::string& value);
  int get_midi_controller_for_port(unsigned long port);
  const DSSI_Program_Descriptor* get_program(unsigned long index);
  void connect_port(unsigned long port, LADSPA_Data* buffer);
  void select_program(unsigned long bank, unsigned long program);
  void run(unsigned long frames);
  void run_synth(unsigned long framecount, 
		 snd_seq_event_t* events, unsigned long eventcount);
  
  /* Here are some functions that return information about the plugin
     so you don't have to mess with the DSSI and LADSPA descriptor structs. */
  const char* label() const;
  const LADSPA_Properties& properties() const;
  const char* name() const;
  const char* maker() const;
  const char* copyright() const;
  unsigned long port_count() const;
  const LADSPA_PortDescriptor& port_descriptor(unsigned long port) const;
  const char* port_name(unsigned long port) const;
  const LADSPA_PortRangeHint& port_range_hint(unsigned long port) const;
  
protected:
  
  void init_midi_map();
  bool load_dssi(const std::string& library, const std::string& label);
  
  std::vector<unsigned long> m_midi_map;
  Glib::Module* m_library;
  LADSPA_Handle m_handle;
  const DSSI_Descriptor* m_dssi;
  bool m_activated;
};


#endif
